#include "caosdb/filestreaming/UploadRequestHandler.h"
#include "caosdb/exceptions.h"                         // for Exception
#include "caosdb/logging.h"                            // for CAOSDB_LOG_ERROR
#include "caosdb/protobuf_helper.h"                    // for get_arena
#include "caosdb/status_code.h"                        // for GENERIC_RPC_E...
#include <algorithm>                                   // for min
#include <boost/filesystem/path.hpp>                   // for operator<<, path
#include <boost/log/core/record.hpp>                   // for record
#include <boost/log/detail/attachable_sstream_buf.hpp> // for basic_ostring...
#include <boost/log/sources/record_ostream.hpp>        // for basic_record_...
#include <boost/preprocessor/seq/limits/enum_256.hpp>  // for BOOST_PP_SEQ_...
#include <boost/preprocessor/seq/limits/size_256.hpp>  // for BOOST_PP_SEQ_...
#include <cstdint>                                     // for uint64_t
#include <exception>                                   // IWYU pragma: keep
// IWYU pragma: no_include <bits/exception.h>
#include <google/protobuf/arena.h>                // for Arena
#include <grpcpp/impl/codegen/async_stream.h>     // for ClientAsyncWr...
#include <grpcpp/impl/codegen/call_op_set.h>      // for WriteOptions
#include <grpcpp/impl/codegen/client_context.h>   // for ClientContext
#include <grpcpp/impl/codegen/completion_queue.h> // for CompletionQueue
#include <grpcpp/impl/codegen/status.h>           // for Status
#include <grpcpp/impl/codegen/status_code_enum.h> // for OK, UNAUTHENT...
#include <iostream>                               // for endl, streamsize
#include <string>                                 // for basic_string
#include <utility>                                // for move

namespace FileExchange {
using caosdb::StatusCode;
using caosdb::exceptions::AuthenticationError;
using caosdb::exceptions::ConnectionError;
using caosdb::exceptions::Exception;
using caosdb::utility::get_arena;
using google::protobuf::Arena;

UploadRequestHandler::UploadRequestHandler(HandlerTag tag,
                                           FileTransmissionService::Stub *stub,
                                           grpc::CompletionQueue *cq,
                                           FileDescriptor file_descriptor)
  : tag_(tag), stub_(stub), cq_(cq),
    request_(Arena::CreateMessage<FileUploadRequest>(get_arena())),
    response_(Arena::CreateMessage<FileUploadResponse>(get_arena())),
    state_(CallState::NewCall), file_descriptor_(std::move(file_descriptor)),
    bytesToSend_(0) {
  this->onNext(true);
}

bool UploadRequestHandler::onNext(bool ok) {
  try {
    if (state_ == CallState::CallComplete) {
      this->handleCallCompleteState();
      return false; // TODO(tf): comment
    } else if (ok) {
      if (state_ == CallState::NewCall) {
        this->handleNewCallState();
      } else if (state_ == CallState::SendingHeader) {
        this->handleSendingHeaderState();
      } else if (state_ == CallState::SendingFile) {
        this->handleSendingFileState();
      } else if (state_ == CallState::ExpectingResponse) {
        this->handleExpectingResponseState();
      }
    } else {
      state_ = CallState::CallComplete;
      rpc_->Finish(&status_, tag_);
    }

    return true;
  } catch (Exception &e) {
    throw;
  } catch (std::exception &e) {
    CAOSDB_LOG_ERROR(logger_name) << "Upload processing error: " << e.what();
  } catch (...) {
    CAOSDB_LOG_ERROR(logger_name)
      << "Upload processing error: unknown exception caught";
  }

  if (state_ == CallState::NewCall) {
    // TODO(tf): comment
    return false;
  }

  ctx_.TryCancel();

  return true;
}

void UploadRequestHandler::cancel() { ctx_.TryCancel(); }

void UploadRequestHandler::handleNewCallState() {
  auto filename = file_descriptor_.local_path;
  fileReader_ = std::make_unique<FileReader>(filename);

  rpc_ = stub_->PrepareAsyncFileUpload(&ctx_, response_, cq_);

  state_ = CallState::SendingHeader;
  rpc_->StartCall(tag_);
}

void UploadRequestHandler::handleSendingHeaderState() {
  auto *tid = request_->mutable_chunk()->mutable_file_transmission_id();
  tid->CopyFrom(*(file_descriptor_.file_transmission_id));

  bytesToSend_ = fileReader_->fileSize();

  if (bytesToSend_ > 0) {
    state_ = CallState::SendingFile;
  } else {
    state_ = CallState::ExpectingResponse;
  }

  rpc_->Write(*request_, tag_);
}

void UploadRequestHandler::handleSendingFileState() {
  const uint64_t DefaultChunkSize = 4 * 1024; // 4K

  auto chunkSize = std::min(DefaultChunkSize, bytesToSend_);

  request_->Clear();
  auto *buffer = request_->mutable_chunk()->mutable_data();
  buffer->resize(chunkSize);

  fileReader_->read(*buffer);
  bytesToSend_ -= chunkSize;

  grpc::WriteOptions writeOptions;
  if (bytesToSend_ > 0) {
    state_ = CallState::SendingFile;
  } else {
    state_ = CallState::ExpectingResponse;
    writeOptions.set_last_message();
  }

  rpc_->Write(*request_, writeOptions, tag_);
}

void UploadRequestHandler::handleExpectingResponseState() {
  state_ = CallState::CallComplete;
  rpc_->Finish(&status_, tag_);
}

void UploadRequestHandler::handleCallCompleteState() {
  switch (status_.error_code()) {
  case grpc::OK: {
    auto bytesSent = fileReader_ ? fileReader_->fileSize() : 0;
    CAOSDB_LOG_INFO(logger_name)
      << "[" << file_descriptor_.local_path
      << "]: upload complete: " << bytesSent << " bytes sent" << std::endl;
  } break;

  case grpc::UNAUTHENTICATED:
    throw AuthenticationError(status_.error_message());
  case grpc::UNAVAILABLE:
    throw ConnectionError(status_.error_message());
  default:
    throw Exception(StatusCode::GENERIC_RPC_ERROR, status_.error_message());
  }
}

} // namespace FileExchange
